import asyncio, traceback, time
from functools import wraps, partial
from copy import copy



def async_wrap(func):
    if asyncio.iscoroutinefunction(func):
        return func
    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        pfunc = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, pfunc)
    return run 


class CoroutinePool:
    def __init__(self, n_workers=10) -> None:
        self.n_workers = n_workers
        self.wait_list = []
        self.ret_list_when_timeout = None


    def _get_task_result(self, task):
        try:
            return task.result()
        except asyncio.CancelledError:
            return traceback.format_exc()
        except:
            pass
        try:
            return task.exception()
        except:
            return traceback.format_exc()


    def add_task(self, func, *args, **kwargs):
        self.wait_list.append((len(self.wait_list), func, args, kwargs))
   

    def execute(self, delay=0, timeout=100, n_workers=None):
        if n_workers is not None:
            self.n_workers = n_workers
        wait_list = copy(self.wait_list)

        async def run():
            start_time = time.time()
            ret_list = [None]*len(wait_list)
            runner = []
            
            for _ in range(1000000000):
                if (time.time()-start_time)>timeout:
                    self.ret_list_when_timeout = ret_list
                    raise ValueError(f'Fail to finish task in {timeout}s')

                # create task
                if len(runner)<self.n_workers and wait_list:
                    idx, func, args, kwargs = wait_list.pop()
                    func = async_wrap(func)
                    task = asyncio.create_task(func(*args, **kwargs))
                    runner.append((idx, task))
                else:
                    # check tasks
                    for i, (idx, task) in enumerate(runner):
                        if not task.done():
                            continue
                        ret_list[idx] = self._get_task_result(task)
                        runner[i] = None
                    runner = [r for r in runner if r]

                    #exit
                    if not runner and not wait_list:
                        break
                    await asyncio.sleep(delay)
            return ret_list

        ret_list = asyncio.run(run())
        self.wait_list = []
        return ret_list
                





class Repository:
    pass


class TaskQueueRepository(Repository):
    def get_request(self):
        raise NotImplementedError
    
    def get_n_requests(self, n):
        raise NotImplementedError
    
    def save_item(self, item):
        raise NotImplementedError
    
    def save_items(self, items):
        raise NotImplementedError
    
    

class TaskProcessRepository(Repository):
    def preprocess(self, item):
        return True, item
    
    def preprocess_batch(self, items):
        return [True]*len(items), items

    def process(self, item):
        raise NotImplementedError
    
    def process_batch(self, items):
        raise NotImplementedError
    
    def postprocess(self, item):
        return True, item
    
    def postprocess_batch(self, items):
        return [True]*len(items), items


import traceback
from retry import retry
from ddd_interface.objects.do import APIGatewayRequestDO
from ddd_interface.objects.lib import deserialize, serialize
from ..domain.repository import TaskQueueRepository



@retry(tries=2, delay=0.5)
def _save_item(gateway_ao, redis_ao, domain, redis_token, task_id, item_str, timeout):
    if gateway_ao:
        assert redis_token
        gateway_request = APIGatewayRequestDO(
            service_name='redis',
            method = 'set_response',
            auth = {'token': redis_token},
            data = {
                'domain': domain,
                'request_id': task_id,
                'response': item_str,
                'timeout': timeout
            }
        )
        gateway_ao.send_request(gateway_request)
    else:
        redis_ao.get_request(
            domain=domain, 
            request_id=task_id, 
            response=item_str, 
            timeout=timeout
        )


class TaskQueueFactory:
    def gen_redis_repo(
        self, 
        domain, 
        request_key, 
        request_class, 
        request_converter,
        item_converter,
        redis_ao=None, 
        gateway_ao = None,
        redis_token = None,
        item_timeout = 86400
    ):
        if not redis_ao and not (gateway_ao and redis_token):
            raise ValueError('redis ao or gateway ao should be defined at least one')
        class RedisTaskQueueRepositoryImpl(TaskQueueRepository):
            @retry(tries=2, delay=0.5)
            def get_task(self):
                if gateway_ao:
                    assert redis_token, 'redis token is not set'
                    gateway_request = APIGatewayRequestDO(
                        service_name='redis',
                        method = 'get_request',
                        auth = {'token': redis_token},
                        data = {
                            'domain': domain,
                            'key': request_key,
                        }
                    )
                    request = gateway_ao.send_request(gateway_request)
                    
                else:
                    request = redis_ao.get_request(domain=domain, key=request_key)

                if request is None:
                    return None
                else:
                    if isinstance(request, dict):
                        request = request['obj']
                    else:
                        request = request.obj
                    request = deserialize(request, request_class)
                    request = request_converter.to_entity(request)
                    return request
            

            def get_n_tasks(self, n):
                requests = []
                for i in range(n):
                    request = self.get_task()
                    if request is None:
                        break
                    else:
                        requests.append(request)
                return requests
                

            def save_item(self, item):
                try:
                    item = item_converter.to_do(item)
                    item_str = serialize(item)
                    task_id = item.task_id
                    timeout = item_timeout
                    _save_item(gateway_ao, redis_ao, domain, redis_token, 
                                task_id, item_str, timeout)
                    return True
                except:
                    traceback.print_exc()
                    return False
            

            def save_items(self, items):
                succeed_symbols = []
                for item in items:
                    succeed = self.save_item(item)
                    succeed_symbols.append(succeed)
                return succeed_symbols
        repo = RedisTaskQueueRepositoryImpl()
        return repo




import time
from copy import copy
from typing import Optional, List
from ..domain.repository import TaskQueueRepository, TaskProcessRepository


class CacheQueue:
    def __init__(self) -> None:
        self.cache = {}
        self.queue = None


    def _next(self):
        while True:
            index = next(self.queue)
            if index in self.cache:
                return index
    

    def _pop(self):
        if self.queue is None:
            self.queue = iter(copy(self.cache))
        try:
            return self.cache[self._next()]
        except StopIteration:
            self.queue = iter(copy(self.cache))
            return None
    

    def _pop_index(self):
        if self.queue is None:
            self.queue = iter(copy(self.cache))
        try:
            return self._next()
        except StopIteration:
            self.queue = iter(copy(self.cache))
            return None
    
    
    def delete(self, index)->bool:
        if index in self.cache:
            del self.cache[index]
            return True
        else:
            return False
    

class SimpleTaskProcessor:
    def __init__(
        self, 
        queue_repo:TaskQueueRepository, 
        process_repo: TaskProcessRepository,
        delay:int = 1
    ) -> None:
        self.queue_repo = queue_repo
        self.process_repo = process_repo
        self.delay = delay
    

    def set_queue_repo(self, queue_repo: TaskQueueRepository):
        self.queue_repo = queue_repo
    

    def set_process_repo(self, process_repo: TaskProcessRepository):
        self.process_repo = process_repo


    def gen_item(self, request):
        raise NotImplementedError
    

    def gen_items(self, requests):
        raise NotImplementedError

    def get_request(self, queue_repo: Optional[TaskQueueRepository]=None):
        queue_repo = queue_repo or self.queue_repo
        return queue_repo.get_task()
    

    def get_requests(self, n=10, queue_repo: Optional[TaskQueueRepository]=None):
        queue_repo = queue_repo or self.queue_repo
        return queue_repo.get_n_tasks(n)
    

    def process_task(self, item, process_repo:Optional[TaskProcessRepository]=None):
        process_repo = process_repo or self.process_repo
        return process_repo.process(item)
    

    def process_tasks(self, items, process_repo:Optional[TaskProcessRepository]=None):
        if process_repo is None:
            process_repo = self.process_repo
        return process_repo.process_batch(items)

    
    def save_item(self, item, queue_repo: Optional[TaskQueueRepository]=None)->bool:
        queue_repo = queue_repo or self.queue_repo
        return queue_repo.save_item(item)
    

    def save_items(self, items, queue_repo: Optional[TaskQueueRepository]=None)->List[bool]:
        queue_repo = queue_repo or self.queue_repo
        return queue_repo.save_items(items)
    

    def process_when_failed(self, item, error_code:str):
        # log or warning by email
        raise NotImplementedError
    

    def process_when_failed_batch(self, items, error_code:str):
        raise NotImplementedError
    

    def postprocess(self, item, process_repo:Optional[TaskProcessRepository]=None)->bool:
        return True, item
    

    def postprocess_batch(
        self, 
        items, 
        process_repo:Optional[TaskProcessRepository]=None
    )->List[bool]:
        return [True]*len(items), items


    def preprocess(self, item, process_repo:Optional[TaskProcessRepository]=None):
        return True, item
    

    def preprocess_batch(
        self, 
        items, 
        process_repo:Optional[TaskProcessRepository]=None
    ):
        return [True]*len(items), items

    
    def process_all(
        self, 
        item, 
        process_repo: TaskProcessRepository, 
        queue_repo: TaskQueueRepository
    ):
        '''处理所有的步骤
        '''
        # preprocess
        succeed, item = self.preprocess(item, process_repo)
        if not succeed:
            self.process_when_failed(item, 'preprocess')
            return 

        # process
        item = self.process_task(item)
        succeed = self.save_item(item, queue_repo)
        if not succeed:
            self.process_when_failed(item, 'save')
            return 

        # postprocess
        succeed, item = self.postprocess(item, process_repo)
        if not succeed:
            self.process_when_failed(item, 'postprocess')
            return

    
    def process_all_batch(
        self, 
        items, 
        process_repo: TaskProcessRepository, 
        queue_repo: TaskQueueRepository
    ):
        '''处理所有步骤，并且是批量处理
        '''
        # preprocess
        succeed_symbols, items = self.preprocess_batch(items, process_repo)
        failed_items = [i for i,s in zip(items, succeed_symbols) if not s]
        if failed_items:
            self.process_when_failed_batch(failed_items, 'preprocess')

        # process
        items = [i for i,s in zip(items, succeed_symbols) if s]
        if not items:
            return
        items = self.process_tasks(items)
        succeed_symbols = self.save_items(items, queue_repo)
        failed_items = [i for i,s in zip(items, succeed_symbols) if not s]
        if failed_items:
            self.process_when_failed_batch(failed_items, 'save')

        # postprocess
        items = [i for i,s in zip(items, succeed_symbols) if s]
        if not items:
            return
        succeed_symbols, items = self.postprocess_batch(items, process_repo)
        failed_items = [i for i,s in zip(items, succeed_symbols) if not s]
        if failed_items:
            self.process_when_failed_batch(failed_items, 'postprocess')

   

    def run(
        self, 
        queue_repo: Optional[TaskQueueRepository]=None, 
        process_repo: Optional[TaskProcessRepository]=None,
        delay=None, 
        n_loop=1000000000
    ):
        queue_repo = queue_repo or self.queue_repo
        process_repo = process_repo or self.process_repo
        delay = delay or self.delay
        for i in range(n_loop):
            time.sleep(delay)
            request = self.get_request(queue_repo)
            if not request:
                continue

            # generate item
            item = self.gen_item(request)
            succeed = self.save_item(item, queue_repo)
            if not succeed:
                self.process_when_failed(item, 'save')
                continue

            # process item
            try:
                self.process_all(item, process_repo)
            except:
                self.process_when_failed(item, 'process_all') 
                
    

    def run_batch(
        self, 
        queue_repo: Optional[TaskQueueRepository]=None, 
        process_repo: Optional[TaskProcessRepository]=None,
        delay=None, 
        n_loop=1000000000
    ):
        queue_repo = queue_repo or self.queue_repo
        process_repo = process_repo or self.process_repo
        delay = delay or self.delay
        for i in range(n_loop):
            time.sleep(delay)
            requests = self.get_requests(queue_repo)
            if not requests:
                continue

            # generate item
            items = self.gen_items(requests)
            succeed_symbols = self.save_items(items, queue_repo)
            failed_items = [i for i,s in zip(items, succeed_symbols) if not s]
            if failed_items:
                self.process_when_failed_batch(failed_items, 'save')
            items = [i for i,s in zip(items, succeed_symbols) if s]
            if not items:
                continue
            try:
                self.process_all_batch(items, process_repo, queue_repo)
            except:
                self.process_when_failed_batch(items, 'process_all')

            




from setuptools import setup, find_packages
version_fn = 'src/ddd_struct/version.py'
with open(version_fn, 'r') as f:
    verstr = f.read()
version = verstr.split('=')[1]
version = version.replace("'", "").strip()


setup(
    name="ddd-struct",
    version=version,
    author="wangziling100",
    author_email="wangziling100@163.com",
    description="Data structures",
    package_dir={"": "src"},
    packages=find_packages(where='src'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=['retry2'],
)